@@ -53,8 +53,8 @@ static int is_incomplete (char *line)
 
 static int validate_config (struct config *cfg, char const **errstr)
 {
-  if (cfg->listen[0] == '\0') {
-    *errstr = "no or empty listen statement";
+  if ((cfg->listen[0] == '\0') && (cfg->geom_listen[0] == '\0')) {
+    *errstr = "no or empty listen statements";
     return -1;
   }
 
@@ -63,6 +63,11 @@ static int validate_config (struct config *cfg, char const **errstr)
     return -1;
   }
 
+  if ((cfg->geom_listen[0] != '\0') && (cfg->geom_port[0] == '\0')) {
+    *errstr = "no or empty geom port statement";
+    return -1;
+  }
+
   if (cfg->num_io_threads == 0) {
     *errstr = "number of workers must not be zero";
     return -1;
@@ -124,16 +129,7 @@ int load_config (char *configfile, struct config *cfg,
   char line[1024], tmp[256];
 
   *err_line = 0;
-
-  /* set default config, prepare mutexes for connection slots */
   memset(cfg, 0, sizeof(*cfg));
-  strcpy(cfg->listen, "/tmp/s3blkdevd.sock");
-  strcpy(cfg->port, "10809");
-  cfg->num_io_threads = 8;
-  cfg->num_s3fetchers = 2;
-  cfg->s3timeout = 10000; // ms
-  cfg->s3ssl = 1;
-  cfg->s3_max_reqs_per_conn = 100;
 
   for (i = 0; i < sizeof(cfg->s3conns)/sizeof(cfg->s3conns[0]); i++) {
     cfg->s3conns[i].sock = -1;
@@ -178,6 +174,8 @@ int load_config (char *configfile, struct config *cfg,
 
     if (sscanf(line, " listen %127s", cfg->listen) ||
         sscanf(line, " port %7[0-9]", cfg->port) ||
+        sscanf(line, " geom_listen %127s", cfg->geom_listen) ||
+        sscanf(line, " geom_port %7[0-9]", cfg->geom_port) ||
         sscanf(line, " workers %hu", &cfg->num_io_threads) ||
         sscanf(line, " fetchers %hu", &cfg->num_s3fetchers) ||
         sscanf(line, " s3maxreqsperconn %hu", &cfg->s3_max_reqs_per_conn) ||
@@ -563,7 +561,7 @@ static ssize_t s3_recv (struct s3connection *conn, void *buffer, size_t buflen,
     ret = read(conn->sock, buffer, buflen);
     if ((ret < 0) && (errno != EAGAIN) && (errno != EWOULDBLOCK) &&
         (errno != EINTR)) {
-      *errstr = gnutls_strerror(ret);
+      *errstr = strerror(errno);
       return -1;
     }
   } else for (;;) {
@@ -708,14 +706,15 @@ static int s3_finish_req (struct s3connection *conn, enum httpverb verb,
       return -1;
 
     readbytes += res;
+    header[readbytes] = '\0';
+
+    if ((body = strstr(header, "\r\n\r\n")) != NULL)
+      break;
+
     if (readbytes >= sizeof(header) - 1) {
       *errstr = "HTTP header too large";
       return -1;
     }
-
-    header[readbytes] = '\0';
-    if ((body = strstr(header, "\r\n\r\n")) != NULL)
-      break;
   }
 
   body += 4;
@@ -747,7 +746,7 @@ static int s3_finish_req (struct s3connection *conn, enum httpverb verb,
   readbytes -= body - header;
   memmove(buffer, body, readbytes);
 
-  if ((verb != HEAD) || (*code != 200)) {
+  if (verb != HEAD) {
     /* receive payload */
     while (readbytes < *contentlen) {
       res = s3_recv(conn, buffer + readbytes,
