@@ -1,7 +1,7 @@
 #ifndef _S3BLKDEV_H
 #define _S3BLKDEV_H
 
-#define S3BLKDEV_VERSION "0.3"
+#define S3BLKDEV_VERSION "0.4"
 
 #include <limits.h>
 #include <time.h>
@@ -83,6 +83,8 @@ struct config {
 
   char listen[128];
   char port[8];
+  char geom_listen[128];
+  char geom_port[8];
 };
 
 int load_config (char *configfile, struct config *cfg,
