@@ -47,6 +47,7 @@
 #define NBD_CMD_WRITE 1
 #define NBD_CMD_DISC 2
 #define NBD_CMD_FLUSH 3
+#define GEOM_MAGIC "GEOM_GATE       "
 
 const char const NBD_INIT_PASSWD[] = { 'N','B','D','M','A','G','I','C' };
 const char const NBD_OPTS_MAGIC[] =  { 'I','H','A','V','E','O','P','T' };
@@ -948,7 +949,7 @@ static void *client_worker (void *arg0)
   if (block_signals() != 0)
     goto ERROR;
 
-  if ((res = pthread_setname_np(pthread_self(), "s3blkdevd:c")) != 0) {
+  if ((res = pthread_setname_np(pthread_self(), "s3blkdevd:nbd")) != 0) {
     logerr("pthread_setname_np(): %s", strerror(res));
     goto ERROR;
   }
@@ -998,6 +999,177 @@ ERROR:
   return NULL;
 }
 
+#if 0
+static int geom_client_worker_loop (struct client_thread_arg *arg)
+{
+  struct __attribute__((packed)) {
+    uint32_t magic;
+    uint32_t type;
+    char handle[8];
+    uint64_t offs;
+    uint32_t len;
+  } req;
+  struct {
+    uint8_t cmd;
+    uint64_t offset;
+    uint32_t length;
+    uint64_t seq;
+    uint16_t error;
+  } __attribute__((packed)) geom_hdr;
+  fd_set rfds;
+  struct timeval timeout;
+  struct io_thread_arg *slot;
+  int res, result = -1;
+
+  FD_ZERO(&rfds);
+  FD_SET(arg->socket, &rfds);
+
+  timeout.tv_sec = 1;
+  timeout.tv_usec = 0;
+
+  res = select(arg->socket + 1, &rfds, NULL, NULL, &timeout);
+  if (res == 0)
+    return 0;
+
+  if (res < 0) {
+    logerr("select(): %s", strerror(errno));
+    goto ERROR;
+  }
+
+  if (!running)
+    goto ERROR;
+
+  slot = find_free_io_worker();
+  if (slot == NULL)
+    goto ERROR;
+
+  if (read_all(arg->socket, &geom_hdr, sizeof(geom_hdr)) != 0)
+    goto ERROR1;
+
+  return 0;
+}
+#endif
+
+static int geom_handshake (struct client_thread_arg *arg)
+{
+  struct {
+    char magic[16];
+    uint16_t version;
+    uint16_t error;
+  } __attribute__((packed)) geom_version;
+  struct {
+    char path[1025];
+    uint64_t flags;
+    uint16_t nconn;
+    uint32_t token;
+  } __attribute__((packed)) geom_cinit;
+  struct {
+    uint8_t flags;
+    uint64_t mediasize;
+    uint32_t sectorsize;
+    uint16_t error;
+  } __attribute__((packed)) geom_sinit;
+
+  if (read_all(arg->socket, &geom_version, sizeof(geom_version)) != 0)
+    return -1;
+
+  if (strncmp(geom_version.magic, GEOM_MAGIC, 16)) {
+    logerr("client %s no GEOM_MAGIC", arg->clientname);
+    return -1;
+  }
+
+  if (ntohs(geom_version.version) != 0) {
+    logerr("client %s unknown geom version %hu",
+           arg->clientname, ntohs(geom_version.version));
+    return -1;
+  }
+
+  geom_version.error = 0;
+
+  if (write_all(arg->socket, &geom_version, sizeof(geom_version)) != 0)
+    return -1;
+
+  if (read_all(arg->socket, &geom_cinit, sizeof(geom_cinit)) != 0)
+    return -1;
+
+  geom_cinit.path[1024] = '\0';
+
+  if ((arg->dev = get_device_by_name(geom_cinit.path)) == NULL) {
+    logerr("client %s unknown device %s", arg->clientname, geom_cinit.path);
+    return -1;
+  }
+
+  geom_sinit.flags = 0;
+  geom_sinit.mediasize = htonll(arg->dev->size);
+  geom_sinit.sectorsize = htonl(512);
+  geom_sinit.error = htons(0);
+
+  if (write_all(arg->socket, &geom_sinit, sizeof(geom_sinit)) != 0)
+    return -1;
+
+  return 0;
+}
+
+static void *geom_client_worker (void *arg0)
+{
+  struct client_thread_arg *arg = (struct client_thread_arg*) arg0;
+  int res;
+
+  if (block_signals() != 0)
+    goto ERROR;
+
+  if ((res = pthread_setname_np(pthread_self(), "s3blkdevd:geom")) != 0) {
+    logerr("pthread_setname_np(): %s", strerror(res));
+    goto ERROR;
+  }
+
+  client_address(arg);
+
+  if (set_socket_options(arg->socket) != 0) {
+    logerr("setsockopt(): %s", strerror(res));
+    goto ERROR;
+  }
+
+  if (geom_handshake(arg) != 0)
+    goto ERROR;
+
+  if ((res = pthread_mutex_init(&arg->socket_mtx, NULL)) != 0) {
+    logerr("pthread_mutex_init(): %s", strerror(res));
+    goto ERROR;
+  }
+
+  arg->cachedir_fd = open(arg->dev->cachedir, O_RDONLY|O_DIRECTORY);
+  if (arg->cachedir_fd < 0) {
+    logerr("open(): %s", strerror(errno));
+    goto ERROR1;
+  }
+
+  syslog(LOG_INFO, "client %s connecting to device %s\n", arg->clientname,
+         arg->dev->name);
+
+#if 0
+  while (geom_client_worker_loop(arg))
+    ;;
+#endif
+
+  syslog(LOG_INFO, "client %s disconnecting from device %s\n", arg->clientname,
+         arg->dev->name);
+
+  if (close(arg->cachedir_fd) != 0)
+    logerr("close(): %s", strerror(errno));
+
+ERROR1:
+  if ((res = pthread_mutex_destroy(&arg->socket_mtx)) != 0)
+    logerr("pthread_mutex_destroy(): %s", strerror(res));
+
+ERROR:
+  if (close(arg->socket) != 0)
+    logerr("close(): %s", strerror(errno));
+
+  free(arg0);
+  return NULL;
+}
+
 static void sigterm_handler (int sig __attribute__((unused)))
 {
   syslog(LOG_INFO, "SIGTERM received, going down...\n");
@@ -1067,6 +1239,9 @@ static int create_listen_socket_inet (char *ip, char *port)
   if (res != 0)
     err(1, "setsockopt()");
 
+  if (listen(listen_socket, 0) != 0)
+    err(1, "listen()");
+
   return listen_socket;
 }
 
@@ -1088,18 +1263,6 @@ static int create_listen_socket_unix (char *ip)
   if (bind(listen_socket, (struct sockaddr*) &sun, sizeof(sun)) != 0)
     err(1, "bind()");
 
-  return listen_socket;
-}
-
-static int create_listen_socket (char *ip, char *port)
-{
-  int listen_socket;
-
-  if (*ip == '/')
-    listen_socket = create_listen_socket_unix(ip);
-  else
-    listen_socket = create_listen_socket_inet(ip, port);
-
   if (listen(listen_socket, 0) != 0)
     err(1, "listen()");
 
@@ -1214,6 +1377,28 @@ static void daemonize ()
     err(1, "chdir(): /");
 }
 
+static int create_worker (int sock, pthread_attr_t *thread_attr,
+                          void *(*worker) (void*))
+{
+  pthread_t thread;
+  struct client_thread_arg *thread_arg;
+
+  thread_arg = malloc(sizeof(*thread_arg));
+  if (thread_arg == NULL)
+    return ENOMEM;
+
+  do {
+    thread_arg->addr_len = sizeof(thread_arg->addr);
+    thread_arg->socket = accept(sock, &thread_arg->addr,
+                                &thread_arg->addr_len);
+  } while ((thread_arg->socket < 0) && (errno == EINTR));
+
+  if (thread_arg->socket < 0)
+    return errno;
+
+  return pthread_create(&thread, thread_attr, worker, thread_arg);
+}
+
 int main (int argc, char **argv)
 {
 #define log_error(fmt, params ...) do { \
@@ -1223,11 +1408,10 @@ int main (int argc, char **argv)
 
   char *configfile = DEFAULT_CONFIGFILE, *pidfile = NULL;
   const char *errstr;
-  int foreground = 0, listen_socket, res;
+  int foreground = 0, listen_socket = -1, geom_listen_socket = -1, res;
   unsigned int errline;
   pthread_attr_t thread_attr;
-  pthread_t thread;
-  struct client_thread_arg *thread_arg;
+  fd_set rfds;
 
   while ((res = getopt(argc, argv, "c:fhp:")) != -1) {
     switch (res) {
@@ -1257,7 +1441,6 @@ int main (int argc, char **argv)
 
   increase_stacksize();
   setup_signals();
-  listen_socket = create_listen_socket(cfg.listen, cfg.port);
   launch_io_workers();
 
   if ((res = pthread_attr_init(&thread_attr)) != 0)
@@ -1266,6 +1449,14 @@ int main (int argc, char **argv)
   if (res != 0)
     errx(1, "pthread_attr_setdetachstate(): %s", strerror(res));
 
+  if (*cfg.listen == '/')
+    listen_socket = create_listen_socket_unix(cfg.listen);
+  else if (*cfg.listen != '\0')
+    listen_socket = create_listen_socket_inet(cfg.listen, cfg.port);
+
+  if (*cfg.geom_listen != '\0')
+    geom_listen_socket = create_listen_socket_inet(cfg.geom_listen,
+                                                   cfg.geom_port);
   if (!foreground) {
     close(0);
     close(1);
@@ -1276,36 +1467,49 @@ int main (int argc, char **argv)
   syslog(LOG_INFO, "starting...\n");
 
   while (running) {
-    thread_arg = malloc(sizeof(*thread_arg));
-    if (thread_arg == NULL) {
-      log_error("%s", "malloc() failed");
+    FD_ZERO(&rfds);
+
+    if (listen_socket > 0)
+      FD_SET(listen_socket, &rfds);
+
+    if (geom_listen_socket > 0)
+      FD_SET(geom_listen_socket, &rfds);
+
+    res = select(MAX(listen_socket, geom_listen_socket) + 1, &rfds, NULL, NULL,
+                 NULL);
+
+    if (res < 0) {
+      if (errno == EINTR)
+        continue;
+
+      log_error("select(): %s", strerror(errno));
       break;
     }
 
-    thread_arg->addr_len = sizeof(thread_arg->addr);
-    thread_arg->socket = accept(listen_socket, &thread_arg->addr,
-                                &thread_arg->addr_len);
-
-    if (thread_arg->socket < 0) {
-      if (errno != EINTR) {
-        log_error("accept(): %s", strerror(errno));
+    if ((listen_socket > 0) && FD_ISSET(listen_socket, &rfds)) {
+      res = create_worker(listen_socket, &thread_attr, &client_worker);
+      if (res != 0) {
+        log_error("create_worker(): %s", strerror(res));
         break;
       }
-
-      free(thread_arg);
-      continue;
     }
 
-    res = pthread_create(&thread, &thread_attr, &client_worker, thread_arg);
-    if (res != 0) {
-      log_error("pthread_create(): %s", strerror(res));
-      break;
+    if ((geom_listen_socket > 0) && FD_ISSET(geom_listen_socket, &rfds)) {
+      res = create_worker(geom_listen_socket, &thread_attr,
+                          &geom_client_worker);
+      if (res != 0) {
+        log_error("create_worker(): %s", strerror(res));
+        break;
+      }
     }
   }
 
   running = 0;
 
-  if (close(listen_socket) != 0)
+  if ((geom_listen_socket > 0) && (close(geom_listen_socket) != 0))
+    log_error("close(): %s", strerror(errno));
+
+  if ((listen_socket > 0) && (close(listen_socket) != 0))
     log_error("close(): %s", strerror(errno));
 
   syslog(LOG_INFO, "waiting for I/O workers...\n");
